import itertools
import re

from polly.core.classes.keyword_match import KeywordMatch


class KeywordMatchHandler:
    def __init__(self, similarity):
        self.similarities = similarity

    def get_keyword_matches(
        self, keywords, value_index, schema_index, **kwargs
    ):
        sk_matches = self.schema_keyword_match_generator(keywords, schema_index)
        vk_matches = self.value_keyword_match_generator(keywords, schema_index)
        return sk_matches + vk_matches

    def value_keyword_match_generator(self, Q, value_index, **kwargs):
        ignored_tables = kwargs.get("ignored_tables", [])
        ignored_attributes = kwargs.get("ignored_attributes", [])

        # Input:  A keyword query Q=[k1, k2, . . . , km]
        # Output: Set of non-free and non-empty tuple-sets Rq

        """
        The tuple-set Rki contains the tuples of Ri that contain all
        terms of K and no other keywords from Q
        """

        # Part 1: Find sets of tuples containing each keyword
        P = {}
        logger.debug("Processing keyword values")
        for keyword in Q:
            if keyword not in value_index:
                continue

            for table in value_index[keyword]:
                if table in ignored_tables:
                    continue

                for attribute, ctids in value_index[keyword][table].items():
                    if (table, attribute) in ignored_attributes:
                        continue

                    vkm = KeywordMatch(
                        table, value_filter={attribute: {keyword}}
                    )
                    P[vkm] = set(ctids)

        # Part 2: Find sets of tuples containing larger termsets
        P = self.disjoint_itemsets(P)

        # Part 3: Ignore tuples
        return list(P)

    def disjoint_itemsets(self, pool):
        # Input: A Set of non-empty tuple-sets for each keyword alone P
        # Output: The Set P, but now including larger termsets (process Intersections)

        """
        Termset is any non-empty subset K of the terms of a query Q
        """

        delayed_removal = {}
        next_stage_pool = {}

        for vkm_i, vkm_j in itertools.combinations(pool, 2):
            if vkm_i.table == vkm_j.table:
                joint_tuples = pool[vkm_i] & pool[vkm_j]

                if len(joint_tuples) > 0:

                    joint_predicates = {}

                    for attribute, keywords in vkm_i.value_filter:
                        joint_predicates.setdefault(attribute, set()).update(
                            keywords
                        )

                    for attribute, keywords in vkm_j.value_filter:
                        joint_predicates.setdefault(attribute, set()).update(
                            keywords
                        )

                    vkm_ij = KeywordMatch(
                        vkm_i.table, value_filter=joint_predicates
                    )
                    next_stage_pool[vkm_ij] = joint_tuples

                    delayed_removal.setdefault(vkm_i, set()).update(
                        joint_tuples
                    )
                    delayed_removal.setdefault(vkm_j, set()).update(
                        joint_tuples
                    )

        for vkm_k in delayed_removal:
            tuples_to_remove = delayed_removal[vkm_k]
            pool[vkm_k].difference_update(tuples_to_remove)
            if len(pool[vkm_k]) == 0:
                del pool[vkm_k]

        if len(next_stage_pool) > 0:
            pool.update(self.disjoint_itemsets(next_stage_pool))
        return pool

    def schema_keyword_match_generator(self, Q, schema_index, **kwargs):
        ignored_tables = kwargs.get("ignored_tables", [])
        ignored_attributes = kwargs.get("ignored_attributes", [])
        threshold = kwargs.get("threshold", 0.99)
        keyword_matches_to_ignore = kwargs.get(
            "keyword_matches_to_ignore", set()
        )

        S = []
        logger.debug("Processing schema matches")
        for keyword in Q:
            for table in schema_index:
                if table in ignored_tables:
                    continue

                for attribute in itertools.chain(
                    "*", schema_index[table].keys()
                ):

                    if (
                        attribute == "id"
                        or (table, attribute) in ignored_attributes
                        or
                        # This line prevents redundant keyword matches
                        attribute == table
                    ):
                        continue

                    # Treat name accordingly
                    table_name = table.replace(".csv", "").split("/")[-1]

                    sim = self.similarities.word_similarity(
                        keyword, table_name, attribute
                    )
                    # logger.debug("similiary : {} threshold: {}".format(sim, threshold))
                    if sim >= threshold:
                        # logger.info(f"found a KWmatch for {keyword} in {table}.{attribute} with score: {sim}")
                        skm = KeywordMatch(
                            table, schema_filter={attribute: {keyword}}
                        )
                        if skm not in keyword_matches_to_ignore:
                            S.append(skm)
        return S

    def filter_kwmatches_by_compound_keywords(
        self, vk_matches, compound_keywords
    ):
        """
        Value-keyword matches which contains only part of a compound_keyword are
        pruned.
        """
        if len(compound_keywords) == 0:
            return vk_matches

        filtered_vk_matches = set()
        for value_keyword_match in vk_matches:
            must_remove = True
            set_a = set(value_keyword_match.keywords())

            for compound_keyword in compound_keywords:
                set_b = set(compound_keyword)

                set_ab = set_a & set_b
                if len(set_ab) == 0 or len(set_ab) == len(set_b):
                    must_remove = False

            if not must_remove:
                filtered_vk_matches.add(value_keyword_match)

        return list(filtered_vk_matches)



from pprint import pprint as pp

from graphviz import Digraph
from IPython.display import display
from sereia.candidate_network import CandidateNetwork
from sereia.keyword_match import KeywordMatch
from sereia.query_match import QueryMatch
from sereia.utils.ordinal import ordinal
from sereia.utils.printmd import printmd
from sereia.utils.shift_tab import shift_tab


class SereiaResult(dict):

    def __init__(self, database_name, index_handler, database_handler, data):
        dict.__init__(self, data)
        self.data = data
        self.database_name = database_name
        self.database_handler = database_handler
        self.index_handler = index_handler

    def plss(self, **kwargs):
        show_str = kwargs.get("show_str", kwargs.get("text", True))
        show_graph = kwargs.get("show_graph", kwargs.get("graph", True))
        show_sql = kwargs.get("show_sql", kwargs.get("sql", True))
        show_df = kwargs.get(
            "show_df", kwargs.get("df", kwargs.get("jnts", True))
        )
        top_k = kwargs.get("top_k", 0)
        head = kwargs.get("head", 5)

        for i, json_cn in enumerate(self.data["candidate_networks"]):
            cjn = CandidateNetwork.from_json_serializable(json_cn)
            printmd("---")
            printmd(f"**{ordinal(i+1)} PLS**:")

            # if show_str:
            #     printmd('---')
            #     print(f'Text:\n{shift_tab(cjn)}')
            if show_graph:
                g = Digraph(
                    graph_attr={"nodesep": "0.2", "ranksep": "0.25"},
                    node_attr={
                        "fontsize": "9.0",
                    },
                    edge_attr={
                        "arrowsize": "0.9",
                    },
                )
                for label, id in cjn.vertices():
                    g.node(id, label=str(label))
                if len(cjn) > 1:
                    for (label_a, id_a), (
                        label_b,
                        id_b,
                    ) in cjn.generate_pipeline():
                        g.edge(id_a, id_b)
                printmd("---")
                print("Graph:")
                display(g)
            if show_df or show_sql:
                base_collection, mongo_query = cjn.get_mongo_query_from_cn(
                    self.database_name,
                    self.index_handler.schema_graph,
                ).build()

                if show_sql:
                    printmd("---")
                    print("Base collection:")
                    print(base_collection, end="\n\n")
                    print("Mongo Query:")
                    pp(mongo_query)

            print()

            if i + 1 >= top_k and top_k > 0:
                break

    def qms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)

        for i, json_qm in enumerate(self.data["query_matches"]):
            qm = QueryMatch.from_json_serializable(json_qm)

            printmd("---")
            printmd(f"**{ordinal(i+1)} QM**:")
            printmd("---")
            print(qm, end="\n")

            if i + 1 >= top_k and top_k > 0:
                break

    def kms(self, **kwargs):
        show_skms = kwargs.get("schema", True)
        show_vkms = kwargs.get("value", True)
        top_k = kwargs.get("top_k", 0)

        self.skms(top_k=top_k)
        self.vkms(top_k=top_k)

    def skms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)
        printmd("---")
        printmd(f"**SKMs**:")
        printmd("---")

        if len(self.data["schema_keyword_matches"]) == 0:
            print("There is no SKM for this keyword query.")
            return
        for i, json_km in enumerate(self.data["schema_keyword_matches"]):
            km = KeywordMatch.from_json_serializable(json_km)
            # print(f'{i+1} KM:')
            print(km)
            if i + 1 >= top_k and top_k > 0:
                break

    def vkms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)
        printmd("---")
        printmd(f"**VKMs**:")
        printmd("---")
        if len(self.data["value_keyword_matches"]) == 0:
            print("There is no VKM for this keyword query.")
            return
        for i, json_km in enumerate(self.data["value_keyword_matches"]):
            km = KeywordMatch.from_json_serializable(json_km)
            print(km)
            if i + 1 >= top_k and top_k > 0:
                break

    def __repr__(self):
        return "SereiaResult"



from pathlib import Path

from polly.index.schema_index import SchemaIndex
from polly.sources.enum import DataSourceType
from polly.sources.handler import SourceHandler
from polly.utils import SchemaGraph


class IndexHandler:
    def __init__(self, config, source_handler: SourceHandler):
        self.config = config
        self.source_handler = source_handler

        self.schema_index = SchemaIndex()
        self.schema_graph = SchemaGraph()

        self.partial_index_count = 0

    def index_exists(self):
        database = self.source_handler.database
        handler_type = self.source_handler.type.value
        dataset_directory = self.config.dataset_directory.format(
            handler_type, database
        )

        file_path = f"{dataset_directory}schema_graph"

        if not Path(file_path).exists():
            return False

        return True

    def create_index(self):
        if not self.config.create_index:
            print("Index Creation is disabled.")
            return

        self.create_schema_graph()

    def create_schema_graph(self):
        fk_constraints = self.source_handler.get_fk_constraints()

        schema_graph: SchemaGraph = SchemaGraph()
        for constraint, values in fk_constraints.items():
            schema_graph.add_fk_constraint(constraint, *values)

        if self.source_handler.type == DataSourceType.DOCUMENT and not len(
            schema_graph
        ):
            schema_graph.add_root(
                self.source_handler.get_collections_names()[0],
            )

        database = self.source_handler.database
        handler_type = self.source_handler.type.value
        dataset_directory = self.config.dataset_directory.format(
            handler_type, database
        )

        schema_graph.persist_to_file(f"{dataset_directory}schema_graph")

    def load_schema_graph(self, **kwargs):
        database = self.source_handler.database
        handler_type = self.source_handler.type.value
        dataset_directory = self.config.dataset_directory.format(
            handler_type, database
        )

        self.schema_graph.load_from_file(f"{dataset_directory}schema_graph")



from .dataframe_sort import (
    sort_dataframe_by_bow_size,
    sort_dataframe_by_token_length,
)
from .get_logger import get_logger
from .graph import Graph
from .keyword_query import KeywordQuery
from .memory import memory_percent, memory_size
from .printmd import printmd
from .schema_graph import SchemaGraph
from .similarity import Similarity
from .tf_iaf import calculate_iaf, calculate_inverse_frequency, calculate_tf
from .tokenizer import Tokenizer



from typing import Any, List

from graphviz import Digraph
from IPython.display import display
from sympy.utilities.misc import ordinal

from polly.core.classes.candidate_network import CandidateNetwork
from polly.core.classes.keyword_match import KeywordMatch
from polly.core.classes.query_match import QueryMatch
from polly.index.handler import IndexHandler
from polly.utils import printmd, shift_tab


class PollyResult:
    def __init__(
        self,
        keyword_query: str,
        database: str,
        keyword_match_count: int,
        query_match_count: int,
        candidate_join_network_count: int,
        keyword_matches: List[Any],
        query_matches: List[Any],
        candidate_join_networks: List[Any],
        generated_query: Any,
        index_handler: IndexHandler,
    ) -> None:
        self.keyword_query = keyword_query
        self.database = database
        self.keyword_match_count = keyword_match_count
        self.query_match_count = query_match_count
        self.candidate_join_network_count = candidate_join_network_count
        self.keyword_matches = keyword_matches
        self.query_matches = query_matches
        self.candidate_join_networks = candidate_join_networks
        self.generated_query = generated_query
        self.index_handler = index_handler

    def cjns(self, **kwargs):
        show_str = kwargs.get("show_str", kwargs.get("text", False))
        show_graph = kwargs.get("show_graph", kwargs.get("graph", True))
        top_k = kwargs.get("top_k", 0)

        for i, json_cn in enumerate(self.candidate_join_networks):
            cjn = CandidateNetwork.from_json_serializable(json_cn)
            printmd("---")
            printmd(f"**{ordinal(i+1)} CJN**:")

            if show_str:
                printmd("---")
                print(f"Text:\n{shift_tab(cjn)}")
            if show_graph:
                g = Digraph(
                    graph_attr={"nodesep": "0.2", "ranksep": "0.25"},
                    node_attr={
                        "fontsize": "9.0",
                    },
                    edge_attr={
                        "arrowsize": "0.9",
                    },
                )
                for label, id in cjn.vertices():
                    g.node(id, label=str(label))
                for (label_a, id_a), (label_b, id_b) in cjn.edges():
                    g.edge(id_a, id_b)
                printmd("---")
                print("Graph:")
                display(g)
            print()

            if i + 1 >= top_k and top_k > 0:
                break

    def qms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)

        for i, json_qm in enumerate(self.query_matches):
            qm = QueryMatch.from_json_serializable(json_qm, self.database)

            printmd("---")
            printmd(f"**{ordinal(i+1)} QM**:")
            printmd("---")
            print(qm, end="\n")

            if i + 1 >= top_k and top_k > 0:
                break

    def kms(self, **kwargs):
        show_skms = kwargs.get("schema", True)
        show_vkms = kwargs.get("value", True)
        top_k = kwargs.get("top_k", 0)

        printmd("---")
        printmd("**KMs**:")
        printmd("---")

        if len(self.keyword_matches) == 0:
            print("There is no KM for this keyword query.")
            return

        for i, json_km in enumerate(self.keyword_matches):
            km = KeywordMatch.from_json_serializable(json_km, self.database)
            # print(f'{i+1} KM:')
            print(km)
            if i + 1 >= top_k and top_k > 0:
                break

        # self.skms(top_k=top_k)
        # self.vkms(top_k=top_k)

    def skms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)
        printmd("---")
        printmd("**SKMs**:")
        printmd("---")

        if len(self.data["schema_keyword_matches"]) == 0:
            print("There is no SKM for this keyword query.")
            return
        for i, json_km in enumerate(self.data["schema_keyword_matches"]):
            km = KeywordMatch.from_json_serializable(json_km, self.database)
            # print(f'{i+1} KM:')
            print(km)
            if i + 1 >= top_k and top_k > 0:
                break

    def vkms(self, **kwargs):
        top_k = kwargs.get("top_k", 0)
        printmd("---")
        printmd("**VKMs**:")
        printmd("---")
        if len(self.data["value_keyword_matches"]) == 0:
            print("There is no VKM for this keyword query.")
            return
        for i, json_km in enumerate(self.data["value_keyword_matches"]):
            km = KeywordMatch.from_json_serializable(json_km, self.database)
            print(km)
            if i + 1 >= top_k and top_k > 0:
                break



# The repr and str methods for the class BabelHash print the underlying key_ids
# used for the dict.
# For a representation within the keys inserted by the setitem method,
# it is recommended to use the methods from pprint or IPython.display modules.


class BabelHash:

    babel = {}
    __slots__ = "_dict"

    def __init__(self, **kwargs):
        self._dict = {}

    def __len__(self):
        return len(self._dict)

    def __getidfromkey__(self, key):
        return BabelHash.babel[key]

    def __getkeyfromid__(self, key_id):
        key = BabelHash.babel[key_id]
        return key

    def __getitem__(self, key):
        key_id = self.__getidfromkey__(key)
        return self._dict[key_id]

    def __setitem__(self, key, value):
        try:
            key_id = BabelHash.babel[key]
        except KeyError:
            key_id = len(BabelHash.babel) + 1

            BabelHash.babel[key] = key_id
            BabelHash.babel[key_id] = key

        self._dict[key_id] = value

    def __delitem__(self, key):
        key_id = self.__getidfromkey__(key)
        self._dict.__delitem__(key_id)

    def __missing__(self, key):
        key_id = self.__getidfromkey__(key)
        return key_id

    def __contains__(self, key):
        try:
            key_id = self.__getidfromkey__(key)
        except KeyError:
            return False

        return self._dict.__contains__(key_id)

    def __iter__(self):
        for key_id in self._dict.keys():
            yield self.__getkeyfromid__(key_id)

    def __repr__(self):
        items = ",".join(f"{key}:{value}" for key, value in self.items())
        return f"<BabelHash {{{items}}}>"

    def keys(self):
        yield from self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self.__getitem__(key)

    def get(self, key):
        value = None
        if key in self:
            value = self.__getitem__(key)
        return value

    def setdefault(self, key, default=None):
        if key not in self:
            self[key] = default
        return self[key]



import pickle
from os import makedirs
from os.path import dirname

from graphviz import Digraph
from IPython import display

from polly.utils import Graph


class SchemaGraph(Graph):
    def __init__(self, graph_dict=None, edges_info=None):
        super().__init__(graph_dict, edges_info)

    def add_fk_constraint(
        self,
        constraint,
        cardinality,
        table,
        foreign_table,
        attribute_mappings,
    ):
        self.add_vertex(table)
        self.add_vertex(foreign_table)

        edge_info = self._edges_info.setdefault((table, foreign_table), {})
        edge_info[constraint] = (cardinality, attribute_mappings)

        self.add_edge(table, foreign_table, edge_info)

    def add_root(self, table):
        self.add_vertex(table)

    def persist_to_file(self, filename):
        data = (self._graph_dict, self._edges_info)
        makedirs(dirname(filename), exist_ok=True)
        with open(filename, mode="wb") as f:
            pickle.dump(data, f)

    def load_from_file(self, filename):
        self._graph_dict = {}
        self._edges_info = {}

        with open(filename, mode="rb") as f:
            data = pickle.load(f)
        self._graph_dict, self._edges_info = data

    def show(self):
        g = Digraph(
            graph_attr={"nodesep": "0.2", "ranksep": "0.25"},
            node_attr={
                "fontsize": "9.0",
            },
            edge_attr={
                "arrowsize": "0.9",
            },
        )
        for id in self.vertices():
            g.node(id, label=str(id.upper()))
        for id_a, id_b in self.edges():
            g.edge(id_a, id_b)
        print("Graph:")
        display(g)

    def __repr__(self):
        if len(self) == 0:
            return "EmptyGraph"
        print_string = [
            "\t" * level + direction + vertex
            for direction, level, vertex in self.leveled_dfs_iter()
        ]

        return "\n".join(print_string)
